"""
IA Parc Inference service
Support for inference of IA Parc models
"""
import os
import io
import asyncio
import uuid
import logging
import logging.config
from typing import Optional, Tuple
import nats
from nats.errors import TimeoutError as NATSTimeoutError
from json_tricks import dumps, loads
from iaparc_inference.config import Config


LEVEL = os.environ.get('LOG_LEVEL', 'INFO').upper()
logging.basicConfig(
    level=LEVEL,
    force=True,
    format="%(name)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
LOGGER = logging.getLogger("Inference")
LOGGER.propagate = True


class IAPListener():
    """
    Inference Listener class
    """

    def __init__(self,
                 callback,
                 batch: Optional[int] = -1,
                 config_path: Optional[str] = "/opt/pipeline/pipeline.json",
                 url: Optional[str] = None,
                 queue: Optional[str] = None,
                 ):
        """
        Constructor
        Arguments:
        - callback:     callback function to process data
                        callback(data: List[bytes], is_batch: bool) -> Tuple[List[bytes], str]
        Optional arguments:
        - batch:        batch size for inference (default: -1)
                        If your model do not support batched self.input, set batch to 1
                        If set to -1, batch size will be determined by the BATCH_SIZE 
                        environment variable
        - config_path:  path to config file (default: /opt/pipeline/pipeline.json)
        - url:          url of inference server (default: None)
                        By default determined by the NATS_URL environment variable,
                        however you can orverride it here
        - self.queue:        name of self.queue (default: None)
                        By default determined by the NATS_self.queue environment variable,
                        however you can orverride it here
        """
        self.callback = callback
        self.batch = batch
        self.config_path = config_path
        self.url = url
        self.queue = queue
        # Init internal variables
        self._dag = None
        self._input = None
        self._output = None

    @property
    def dag(self) -> Config:
        """ Input property """
        if self._dag is None:
            self._dag = Config(self.config_path)
        return self._dag

    @property
    def input(self) -> str:
        """ Input property """
        if self._input is None:
            self._input = self.dag.input
        return self._input

    @input.setter
    def input(self, value: str):
        self._input = value

    @property
    def output(self) -> str:
        """ Input property """
        if self._output is None:
            self._output = self.dag.output
        return self._output

    @output.setter
    def output(self, value: str):
        self._output = value

    def run(self):
        """
        Run inference service
        """
        self.url = self.url or os.environ.get(
            "NATS_URL", "nats://localhost:4222")
        self.queue = self.queue or os.environ.get(
            "NATS_self.queue", "inference")
        self.batch = self.batch or int(os.environ.get("BATCH_SIZE", 1))
        asyncio.run(self._run_async())

    async def _run_async(self):
        """ Start listening to NATS messages
        url: NATS server url
        self.queue_in: input self.queue
        self.queue_out: output self.queue
        batch_size: batch size
        """
        nc = await nats.connect(self.url)
        js = nc.jetstream()
        queue_in = self.queue + "." + self.input
        queue_out = self.queue + "." + self.output
        print("Listening on self.queue:", queue_in)
        print("Sending to self.queue:", queue_out)
        l = len(queue_in+".")
        sub_in = await js.subscribe(queue_in+".>",
                                    queue=self.queue+"-"+self.input,
                                    stream=self.queue)
        data_store = await js.object_store(bucket=self.queue+"-data")

        async def get_data(msg):
            uid = msg.subject[l:]
            source = msg.headers.get("DataSource", "")
            data = None
            if source == "json":
                data = loads(msg.data.decode())
            elif source == "object_store":
                obj_res = await data_store.get(msg.data.decode())
                data = obj_res.data
            elif source == "file":
                file = io.BytesIO()
                obj_res = await data_store.get(msg.data.decode(), file)
                file.read()
            else:
                data = msg.data

            return (uid, source, data)

        async def send_reply(uid, source, data, error=""):
            _out = queue_out + "." + uid
            breply = b''
            if data is not None:
                if source == "json":
                    breply = str(dumps(data)).encode()
                elif source == "object_store":
                    uid = str(uuid.uuid4())
                    breply = uid.encode()
                    await data_store.put(uid, data)
            await js.publish(_out, breply, headers={"ProcessError": error, "DataSource": source})

        async def handle_msg(msgs, is_batch: bool):
            if is_batch:
                batch, uids, sources = zip(*[await get_data(msg) for msg in msgs])
                batch = list(batch)
                reply, err = self._process_data(batch, is_batch)
                if not err:
                    for data, uid, source in zip(reply, uids, sources):
                        await send_reply(uid, source, data)
                    return
                for uid, source in zip(uids, sources):
                    await send_reply(uid, source, None, err)
                return

            for msg in msgs:
                print("handle msg", msg.subject)
                uid, source, data = await get_data(msg)
                reply, err = self._process_data([data], is_batch)
                if err:
                    await send_reply(uid, source, reply, err)
                else:
                    await send_reply(uid, source, reply[0])
                return

        async def term_msg(msgs):
            for msg in msgs:
                await msg.ack()

        # Mark as running
        os.system("touch /tmp/running")
        # Fetch and ack messagess from consumer.
        while True:
            try:
                pending_msgs = sub_in.pending_msgs
                if self.batch == 1 or pending_msgs == 0:
                    print("waiting for messages")
                    msg = await sub_in.next_msg(timeout=600)
                    await asyncio.gather(
                        handle_msg([msg], False),
                        term_msg([msg])
                    )
                else:
                    if pending_msgs >= self.batch:
                        _batch = self.batch
                    else:
                        _batch = pending_msgs
                    msgs = []
                    done = False
                    i = 0
                    while not done:
                        try:
                            msg = await sub_in.next_msg(timeout=0.01)
                            msgs.append(msg)
                        except TimeoutError:
                            done = True
                        i += 1
                        if i == _batch:
                            done = True
                        p = sub_in.pending_msgs
                        if p == 0:
                            done = True
                        elif p < _batch - i:
                            _batch = p + i

                    await asyncio.gather(
                        handle_msg(msgs, True),
                        term_msg(msgs)
                    )
            except NATSTimeoutError:
                continue
            except TimeoutError:
                continue
            except Exception as e: # pylint: disable=W0703
                LOGGER.error("Fatal error message handler: %s",
                             str(e), exc_info=True)
                break
        await nc.close()

    def _process_data(self, requests: list, is_batch: bool = False) -> Tuple[list, str | None]:
        """
        Process data
        Arguments:
        - requests:   list of data to process
        - is_batch:   is batched data
        Returns:
        - Tuple[List[bytes], str]:  list of processed data and error message
        """
        try:
            LOGGER.debug("handle request")
            result = self.callback(requests)
            if is_batch:
                if not isinstance(result, list):
                    return [], "batch reply is not a list"
                if len(requests) != len(result):
                    return [], "batch reply has wrong size"
            return result, None
        except ValueError:
            LOGGER.error("Fatal error message handler", exc_info=True)
            return [], "Wrong input"
        except Exception as e: # pylint: disable=W0703
            LOGGER.error("Fatal error message handler", exc_info=True)
            return [], str(e)



"""Config reader class"""
import logging.config
import os
import json
from json.decoder import JSONDecodeError
from yaml import Loader, dump, load


LOGGER = logging.getLogger("Config")

class Config(dict):
    """
    Config reader class
    """
    def __init__(self, data=None):
        super(Config, self).__init__()
        if data:
            if isinstance(data, dict):
                self.__update(data, {})
            elif isinstance(data, str):
                filename = os.path.basename(data)
                ext = os.path.splitext(filename)[1]
                self.__path = data
                self.__ext = ext
                if ext == "json":
                    self.__update(self.load_json(data), {})
                elif ext == "yaml" or ext == "yml":
                    self.__update(self.load_yaml(data), {})
                else:
                    try:
                        self.__update(self.load_json(data), {})
                    except (JSONDecodeError,TypeError):
                        self.__update(self.load_yaml(data), {})
            else:
                raise ValueError("Unknown data format")

    @staticmethod
    def dump_yaml(data, file_name):
        '''Dump data to yaml file'''
        to_dump = data.copy()
        del to_dump['_Config__path']
        del to_dump['_Config__ext']
        with open(f"{file_name}", "w", encoding="utf-8") as f:
            dump(to_dump, f)

    @staticmethod
    def dump_json(data, file_name):
        '''Dump data to json file'''
        to_dump = data.copy()
        del to_dump['_Config__path']
        del to_dump['_Config__ext']
        with open(f"{file_name}", "w", encoding="utf-8") as f:
            f.writelines(json.dumps(to_dump, indent=4))

    def save(self):
        '''Save config to file'''
        try:
            if self.__ext.lower() == ".json":
                self.save_to_json(self.__path)
            elif self.__ext.lower() == ".yaml" or self.__ext.lower() == ".yml":
                self.save_to_yaml(self.__path)
            else:
                LOGGER.error("Cannot save file, unknown extenstion %s", self.__ext)
        except Exception:
            LOGGER.error("Cannot save config", exc_info=True)

    def save_to_json(self, filename):
        '''Save config to json file'''
        self.dump_json(self, filename)

    def save_to_yaml(self, filename):
        '''Save config to yaml file'''
        self.dump_yaml(self, filename)

    @staticmethod
    def load_json(config):
        '''Load json file'''
        with open(config, "r", encoding="utf-8") as f:
            data = json.load(f)
        return data

    @staticmethod
    def load_yaml(config):
        '''Load yaml file'''
        with open(config, "r", encoding="utf-8") as f:
            data = load(f, Loader=Loader)
        return data

    def new(self, data):
        '''Create new config from data'''
        self.__update(data, {})

    def load(self, data, did):
        """load methode"""
        self.__update(data, did)

    def __update(self, data, did):
        dataid = id(data)
        did[dataid] = self
        for k in data:
            dkid = id(data[k])
            if dkid in did.keys():
                self[k] = did[dkid]
            elif isinstance(data[k], Config):
                self[k] = data[k]
            elif isinstance(data[k], dict):
                obj = Config()
                obj.load(data[k], did)
                self[k] = obj
                obj = None
            elif isinstance(data[k], list) or isinstance(data[k], tuple):
                self[k] = self._add_list(data[k], did)
            else:
                self[k] = data[k]

    def _add_list(self, data, did):
        lst = []
        for l in data:
            if isinstance(l, dict):
                obj = Config()
                obj.load(l, did)
                lst.append(obj)
                obj = None
            elif isinstance(l, list) or isinstance(l, tuple):
                lst.append(self._add_list(l, did))
            else:
                lst.append(l)
        if isinstance(data, tuple):
            lst = tuple(lst)
        return lst

    def __getattr__(self, key):
        return self.get(key, None)

    def __setattr__(self, key, value):
        if isinstance(value, dict):
            self[key] = Config(value)
        else:
            self[key] = value

    def has_key(self, k):
        """ returns True if key is present in the config"""
        if k in self.keys():
            return True
        else:
            return False

    def update(self, *args):
        for obj in args:
            for k in obj:
                if isinstance(obj[k], dict):
                    self[k] = Config(obj[k])
                else:
                    self[k] = obj[k]
        return self

    def merge(self, *args):
        """ merges the config with one or more configs"""
        for obj in args:
            for k in obj:
                if k in self.keys():
                    if isinstance(self[k], list) and isinstance(obj[k], list):
                        self[k] += obj[k]
                    elif isinstance(self[k], list):
                        self[k].append(obj[k])
                    elif isinstance(obj[k], list):
                        self[k] = [self[k]] + obj[k]
                    elif isinstance(self[k], Config) and isinstance(obj[k], Config):
                        self[k].merge(obj[k])
                    elif isinstance(self[k], Config) and isinstance(obj[k], dict):
                        self[k].merge(obj[k])
                    else:
                        self[k] = [self[k], obj[k]]
                else:
                    if isinstance(obj[k], dict):
                        self[k] = Config(obj[k])
                    else:
                        self[k] = obj[k]
        return self

    def replace_variables(self, variables):
        """ replaces all variables in the config with the given variables"""
        for k, obj in self.items():
            if isinstance(obj, Config):
                obj.replace_variables(variables)
            elif isinstance(obj, str):
                self[k] = obj.format(**variables)
            elif isinstance(obj, list) or isinstance(obj, tuple):
                self[k] = self.replace_in_list(obj, variables)

    def replace_in_list(self, obj, variables):
        """ replaces all variables in the list with the given variables"""
        for i, entry in enumerate(obj):
            if isinstance(entry, Config):
                entry.replace_variables(variables)
            elif isinstance(entry, str):
                obj[i] = entry.format(**variables)
            elif isinstance(obj, list) or isinstance(obj, tuple):
                self.replace_in_list(obj, variables)
        return obj



# def test01():
#     """Test Config class"""
#     class UObject(Config):
#         """Test class"""

#     obj = Config({1: 2})
#     d = {}
#     d.update({
#         "a": 1,
#         "b": {
#             "c": 2,
#             "d": [3, 4, 5],
#             "e": [[6, 7], (8, 9)],
#             "self": d,
#         },
#         1: 10,
#         "1": 11,
#         "obj": obj,
#     })
#     x = UObject(d)
#     print(x)

#     assert x.a == x["a"] == 1
#     assert x.b.c == x["b"]["c"] == 2
#     assert x.b.d[0] == 3
#     assert x.b.d[1] == 4
#     assert x.b.e[0][0] == 6
#     assert x.b.e[1][0] == 8
#     assert x[1] == 10
#     assert x["1"] == 11
#     assert x[1] != x["1"]
#     assert id(x) == id(x.b.self.b.self) == id(x.b.self)
#     assert x.b.self.a == x.b.self.b.self.a == 1

#     x.x = 12
#     assert x.x == x["x"] == 12
#     x.y = {"a": 13, "b": [14, 15]}
#     assert x.y.a == 13
#     assert x.y.b[0] == 14


# def test02():
#     """Test Config class"""
#     x = Config({
#         "a": {
#             "b": 1,
#             "c": [2, 3]
#         },
#         1: 6,
#         2: [8, 9],
#         3: 11,
#     })
#     y = Config({
#         "a": {
#             "b": 4,
#             "c": [5]
#         },
#         1: 7,
#         2: 10,
#         3: [12, 13],
#     })
#     z = {
#         3: 14,
#         2: 15,
#         "a": {
#             "b": 16,
#             "c": 17,
#         }
#     }
#     x.merge(y, z)
#     assert 2 in x.a.c
#     assert 3 in x.a.c
#     assert 5 in x.a.c
#     assert 1 in x.a.b
#     assert 4 in x.a.b
#     assert 8 in x[2]
#     assert 9 in x[2]
#     assert 10 in x[2]
#     assert 11 in x[3]
#     assert 12 in x[3]
#     assert 13 in x[3]
#     assert 14 in x[3]
#     assert 15 in x[2]
#     assert 16 in x.a.b
#     assert 17 in x.a.c


# def test03():
#     """Test Config class"""
#     variables = {"name": "robert", "age": 23}
#     x = Config({
#         "person": {
#             "name": "{name}",
#             "age": "{age}"
#         },
#         "desc": "{name} is of age {age}",
#         "lst1": ["{name}", "{age}"],
#         "lst2": ["{name}", "{age}"]
#     })
#     x.replace_variables(variables)
#     print(x)

# if __name__ == '__main__':
#     test01()
#     test02()
#     test03()
